from unikube.cli.console.logger import error, info


def exit_login_required():
    info("You need to login (again). Please run 'unikube login' and try again.")
    exit(1)


def exit_generic_error():
    error("Something went wrong!")
    exit(1)



import unikube.cli.console as console


def container_list(data):
    if len(data.spec.containers) <= 1:
        return None

    container = console.list(
        message="Please select a container",
        message_no_choices="No container is running.",
        choices=[c.name for c in data.spec.containers],
    )
    if container is None:
        return None

    return container



import sys

import click
from click_didyoumean import DYMGroup

import unikube.cli.console as console
from unikube.cli import app as app_cmd
from unikube.cli import auth as auth_cmd
from unikube.cli import context as context_cmd
from unikube.cli import deck as deck_cmd
from unikube.cli import orga as orga_cmd
from unikube.cli import project as project_cmd
from unikube.cli import system as system_cmd
from unikube.cli import unikube as unikube_cmd
from unikube.completion.completion import render_completion_script
from unikube.context import ClickContext
from unikube.helpers import compare_current_and_latest_versions

version = sys.version_info
if version.major == 2:
    console.error("Python 2 is not supported for Unikube. Please upgrade python.", _exit=True)


@click.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_context
def cli(ctx, **kwargs):
    """
    The Unikube CLI provides several command groups to manage all required aspects to develop cloud native
    software on a Kubernetes-based environment.

    There are a couple of shortcut commands directly available from here.
    """
    ctx.obj = ClickContext()


# click -----
@click.command()
def version():
    """
    Check unikube version.
    """
    version = compare_current_and_latest_versions()
    if version is None:
        console.error("Could not determine version.")

    console.info(f"unikube, version {version}")


cli.add_command(version)
cli.add_command(unikube_cmd.ps)


@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_obj
def system(ctx):
    """
    The ``system`` command group includes commands to manage system dependencies on your local machine.
    Using :ref:`reference/system:install` and :ref:`reference/system:verify` you can install all necessary
    dependencies for Unikube and verify their versions.
    """


@click.command()
@click.argument("shell", required=True)
def completion(shell):
    """
    Generate tab completion script for a given shell.
    Supported shells: bash.
    """

    render_completion_script(cli, shell)


# system
system.add_command(system_cmd.install)
system.add_command(system_cmd.verify)
system.add_command(completion)


# organization
@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_obj
def orga(ctx):
    """
    Every registered user can belong to one or multiple organisations and can get authorized for the projects of that
    organisation. This command group manages information about your organisations.
    You can see all organizations you belong to with the :ref:`list command<reference/orga:list>`. It presents a
    tabular view of organisations with ``id`` and ``name``. The :ref:`info command<reference/orga:info>` can be used to
    get more detailed information about a particular organisation. This command displays the ``id``, ``title`` and the
    optional description of the organisation. The organisation belongs to the group of selection commands, thus it gives
    three possible options:

        1. you can either manually enter the ``organization_id`` as an optional argument

        2. you can have a context already set with ``organization_id``, then the info for the set organisation will be
           displayed

        3. if none of the above options is specified, you will be prompted with the selection of all possible
           organisations you have access to.

    """


orga.add_command(orga_cmd.list)
orga.add_command(orga_cmd.info)


# project
@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_obj
def project(ctx):
    """
    Manage your projects.
    """


project.add_command(project_cmd.list)
project.add_command(project_cmd.info)
project.add_command(project_cmd.up)
project.add_command(project_cmd.down)
project.add_command(project_cmd.delete)
project.add_command(project_cmd.prune)


# deck
@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_obj
def deck(ctx):
    """
    Manage all decks you have access to.
    """


deck.add_command(deck_cmd.list)
deck.add_command(deck_cmd.info)
deck.add_command(deck_cmd.install)
deck.add_command(deck_cmd.uninstall)
deck.add_command(deck_cmd.ingress)


# application
@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_obj
def app(ctx):
    """
    Manage your applications.
    """


app.add_command(app_cmd.info)
app.add_command(app_cmd.list)
app.add_command(app_cmd.shell)
app.add_command(app_cmd.switch)
app.add_command(app_cmd.logs)
app.add_command(app_cmd.env)
app.add_command(app_cmd.exec)
app.add_command(app_cmd.update)


# authentication
@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
def auth():
    """
    The authentication command group unites all subcommands for managing Unikube's authentication process. Besides the
    standard :ref:`reference/auth:login` and :ref:`reference/auth:logout` commands, you can check your current
    authentication status by using :ref:`reference/auth:status` command.
    A valid login state is required for most of the unikube CLI commands.
    """


auth.add_command(auth_cmd.login)
auth.add_command(auth_cmd.logout)
auth.add_command(auth_cmd.status)


# context
@cli.group(cls=DYMGroup, max_suggestions=2, cutoff=0.5)
@click.pass_obj
def context(ctx):
    """
    The ``context`` command group enables you to modify the local context.
    You can :ref:`reference/context:set` and :ref:`reference/context:remove` the organization, project
    and deck context. Use :ref:`reference/context:show` to show the current context.
    """


context.add_command(context_cmd.set)
context.add_command(context_cmd.remove)
context.add_command(context_cmd.show)


# shortcut
# -> include auth check in functions if required
cli.add_command(auth_cmd.login)
cli.add_command(auth_cmd.logout)
cli.add_command(project_cmd.up)
cli.add_command(deck_cmd.install)
cli.add_command(app_cmd.shell)



import re
import sys
from pathlib import Path
from urllib.parse import urljoin

import click_spinner
import pkg_resources
import requests
from requests import HTTPError, Session

import unikube.cli.console as console
from unikube import settings
from unikube.authentication.authentication import TokenAuthentication
from unikube.context import ClickContext
from unikube.graphql_utils import EnvironmentType
from unikube.local.providers.types import K8sProviderType
from unikube.local.system import Telepresence


def get_requests_session(access_token) -> Session:
    session = requests.Session()
    session.headers.update({"Content-type": "application/json", "Authorization": "Bearer " + str(access_token)})
    return session


def download_specs(access_token: str, environment_id: str):
    session = get_requests_session(access_token=access_token)

    manifest_url = urljoin(settings.MANIFEST_DEFAULT_HOST, environment_id)
    with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
        response = session.get(manifest_url)
    response.raise_for_status()

    manifest = response.json()
    return manifest


def download_manifest(deck: dict, authentication: TokenAuthentication, access_token: str, environment_index: int = 0):
    try:
        environment_id = deck["environment"][environment_index]["id"]
        console.info("Requesting manifests. This process may take a few seconds.")
        manifest = download_specs(
            access_token=access_token,
            environment_id=environment_id,
        )
    except HTTPError as e:
        project_id = deck["project"]["id"]
        if e.response.status_code == 404:
            console.warning(
                "This deck does potentially not specify a valid Environment of type 'local'. "
                f"Please go to https://app.unikube.io/project/{project_id}/decks "
                f"and save a valid values path."
            )
            exit(1)
        elif e.response.status_code == 403:
            console.warning("Refreshing access token")
            environment_id = deck["environment"][environment_index]["id"]
            response = authentication.refresh()
            if not response["success"]:
                console.exit_login_required()

            access_token = response["response"]["access_token"]
            try:
                manifest = download_specs(
                    access_token=access_token,
                    environment_id=environment_id,
                )
            except HTTPError as e:
                console.warning(f"Even after refreshing access token download specs fails with {e}")
                exit(1)
        else:
            console.error("Could not load manifest: " + str(e), _exit=True)

    return manifest


# environment
def environment_type_from_string(environment_type: str):
    try:
        environment_type = EnvironmentType(environment_type)
    except Exception as e:
        console.debug(e)
        environment_type = None

    return environment_type


def check_environment_type_local_or_exit(deck: dict, environment_index: int = 0):
    if (
        environment_type_from_string(environment_type=deck["environment"][environment_index]["type"])
        != EnvironmentType.LOCAL
    ):
        console.error("This deck cannot be installed locally.", _exit=True)


def check_running_cluster(ctx: ClickContext, cluster_provider_type: K8sProviderType.k3d, project_instance: dict):
    for cluster_data in ctx.cluster_manager.get_all():
        cluster = ctx.cluster_manager.select(cluster_data=cluster_data, cluster_provider_type=cluster_provider_type)
        if cluster.exists() and cluster.ready():
            if cluster.name == project_instance["title"] and cluster.id == project_instance["id"]:
                Telepresence(cluster.storage.get()).start()
                console.info(f"Kubernetes cluster for '{cluster.display_name}' is already running.", _exit=True)
            else:
                console.error(
                    f"You cannot start multiple projects at the same time. Project {cluster.name} ({cluster.id}) is "
                    f"currently running. Please run 'unikube project down {cluster.id}' first and "
                    f"try again.",
                    _exit=True,
                )


def compare_current_and_latest_versions():
    try:
        current_version = None
        try:
            path = Path(__file__).parent / "../VERSION"
            with path.open("r") as f:
                current_version = f.read()
        except (FileNotFoundError, PermissionError):
            console.debug("Could not read current version.")

        if not current_version:
            dist = pkg_resources.working_set.by_key.get("unikube")
            if dist:
                current_version = dist.version

        release = requests.get("https://api.github.com/repos/unikubehq/cli/releases/latest")
        if release.status_code == 403:
            console.info("Versions cannot be compared, as API rate limit was exceeded")
            return None
        latest_release_version = release.json()["tag_name"].replace("-", ".")
        if current_version != latest_release_version:
            console.info(
                f"You are using unikube version {current_version}; however, version {latest_release_version} is "
                f"available."
            )

        return current_version
    except pkg_resources.DistributionNotFound as e:
        console.warning(f"Version of the package could not be found: {e}")
    except Exception:
        import traceback

        console.info(f"Versions cannot be compared, because of error {traceback.format_exc()}")


def compare_decorator(f):
    compare_current_and_latest_versions()



from tests.login_testcase import LoginTestCase
from unikube.cli import app
from unikube.commands import ClickContext


def check():
    """Function used to mock check function"""
    pass


class AppTestCase(LoginTestCase):
    def test_list(self):
        obj = ClickContext()
        obj.auth.check = check
        result = self.runner.invoke(
            app.list,
            obj=obj,
        )
        assert result.exit_code == 1

    def test_shell_invalid_arguments(self):
        obj = ClickContext()
        obj.auth.check = check
        result = self.runner.invoke(
            app.shell,
            [
                "test",
                "--organization",
                "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
                "--project",
                "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
                "--deck",
                "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
            ],
            obj=obj,
        )
        assert "[ERROR] Something went wrong!\n" in result.output



import fnmatch
import uuid
from functools import lru_cache
from typing import KeysView, List, Optional, Union

import jwt
from pydantic import BaseModel
from retrying import retry

from unikube import settings
from unikube.authentication.authentication import IAuthentication
from unikube.authentication.types import AuthenticationData
from unikube.cli import console


class KeycloakPermissionData(BaseModel):
    scopes: Optional[List[str]]
    rsid: str
    rsname: str


class KeycloakPermissions:
    def __init__(self, authentication: IAuthentication):
        self.authentication = authentication

    def _permission_data(self):
        # verify
        response = self.authentication.verify_or_refresh()
        if not response:
            console.exit_login_required()

        # get authentication_data
        authentication_data = self.authentication.general_data.authentication

        # check for requesting_party_token
        if not authentication_data.requesting_party_token:
            raise Exception("Requesting Party Token (RPT) required.")

        # decode requesting_party_token
        requesting_party_token = self._decode_requesting_party_token(
            requesting_party_token=authentication_data.access_token
        )

        # convert
        permission_data = KeycloakPermissions._convert(requesting_party_token["authorization"]["permissions"])

        return permission_data

    def _decode_requesting_party_token(self, requesting_party_token: str) -> dict:
        # decode
        try:
            token = jwt.decode(
                requesting_party_token,
                algorithms=["RS256"],
                audience=settings.TOKEN_AUDIENCE,
                options={"verify_signature": False},
            )
        except Exception as e:
            console.debug(e)
            raise Exception("Requesting Party Token (RPT) could not be decoded.")

        return token

    @staticmethod
    def _convert(permissions: dict) -> List[KeycloakPermissionData]:
        keycloak_permission_list = []
        for permission_dict in permissions:
            keycloak_permission = KeycloakPermissionData(**permission_dict)
            keycloak_permission_list.append(keycloak_permission)

        return keycloak_permission_list

    @lru_cache(10)
    def get_permissions_by_scope(self, scope: str) -> List[KeycloakPermissionData]:
        """
        Return a list of resources with the given scope. Supports to filter with wildcards
        e.g. organization:*.
        """
        permission_data = self._permission_data()

        results = []
        for permission in permission_data:
            if permission.scopes:
                # 'scopes': ['organization:view', 'organization:edit']
                matched = fnmatch.filter(permission.scopes, scope)
                if matched:
                    results.append(permission)

        return results



import unittest

from click.testing import CliRunner

from unikube.commands import ClickContext, completion


class CompletionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()

    def test_no_cluster(self):
        result = self.runner.invoke(
            completion,
            ["bash"],
            obj=ClickContext(),
        )
        self.assertEqual(result.exit_code, 0)



from typing import Optional

from unikube.local.providers.types import K8sProviderData


class K3dData(K8sProviderData):
    api_port: Optional[str] = None
    publisher_port: Optional[str] = None
    kubeconfig_path: Optional[str] = None



# -*- coding: utf-8 -*-
from typing import List, Optional

from tinydb import Query
from utils.client import GQLQueryExecutor, get_requests_session

from unikube import settings


class ConfigManager:

    DB = None
    MISC = settings.MISC
    QUERY = Query()

    def set_active(self, _id, slug, **kwargs) -> dict:
        self.DB.update({"cli_active": False})
        obj = self.DB.get(self.QUERY.id == _id)
        if obj:
            obj.update(kwargs)
            obj["cli_active"] = True
            self.DB.write_back([obj])
        else:
            # object gets created for the very first time
            obj = {"id": _id, "name": slug, "cli_active": True}
            obj.update(kwargs)
            self.DB.insert(obj)
        return obj

    def update_active(self, obj):
        db_obj = self.DB.get(self.QUERY.id == obj["id"])
        db_obj.update(obj)
        self.DB.write_back([db_obj])

    def get_active(self) -> Optional[dict]:
        obj = self.DB.get(self.QUERY.cli_active == True)  # noqa
        return obj

    def get_all(self) -> List[dict]:
        return self.DB.all()

    def delete(self, _id):
        self.DB.remove(self.QUERY.id == _id)


class AppManager(ConfigManager):
    DB = settings.config.table("apps")

    def unset_app(self):
        self.DB.update({"cli_active": False})


class ProjectManager(ConfigManager):
    DB = settings.config.table("projects")
    APP_MGR = AppManager()

    def unset_project(self):
        self.DB.update({"cli_active": False})
        self.APP_MGR.unset_app()


class AllProjects(GQLQueryExecutor):
    query = """
    {
      projects(organizationId: "") {
        id
        slug
        description
      }
    }
    """
    key = "projects"


class ProjectInfo(GQLQueryExecutor):
    query = """
    {
      project(id: "$id") {
        id
        slug
        description
        organization {
          name
        }
        specRepository
        created
      }
    }
    """
    key = "project"


class ProjectApps(GQLQueryExecutor):
    query = """
    {
      project(id: "$id") {
        applications {
          id
          slug
          description
          namespace
          environment(level:"local"){
               specsUrl
          }
        }
      }
    }
    """
    key = "project"

    def get_data(self, **kwargs):
        data = self._query(**kwargs)["applications"]
        if "filter" in kwargs:
            _filter = kwargs["filter"]
            result = []
            if type(_filter) == list:
                for d in data:
                    [d.pop(x, None) for x in _filter]
                    result.append(d)
                return result
            else:
                for d in data:
                    d.pop(_filter)
                    result.append(d)
                return result
        return data


class AppSpecs(GQLQueryExecutor):
    query = """
    {
        applications(id: "$id") {
            namespace
            environment(level:"local"){
               specsUrl
            }
        }
    }
    """

    key = "applications"


class Deployments(GQLQueryExecutor):
    query = """
    {
        application(id: "$id") {
            namespace
            deployments(level: "local") {
                id
                slug
                description
                ports
                isSwitchable
            }
        }
    }
    """

    key = "application"


def download_specs(url):
    session = get_requests_session()
    r = session.get(settings.DEFAULT_UNIKUBE_GRAPHQL_HOST + url)
    if r.status_code == 200:
        return r.json()
    raise Exception(f"access to K8s specs failed (status {r.status_code})")



from typing import List, Union

import unikube.cli.console as console
from unikube import settings
from unikube.local.providers.abstract_provider import AbstractK8sProvider
from unikube.local.providers.factory import kubernetes_cluster_factory
from unikube.local.providers.types import K8sProviderData, K8sProviderType
from unikube.storage.local_storage import LocalStorage


class K8sClusterManager(LocalStorage):
    table_name = "clusters"
    pydantic_class = K8sProviderData

    def get_all(self) -> List[K8sProviderData]:
        cluster_list = []
        for item in self.database.table.all():
            try:
                cluster_data = K8sProviderData(**item)
                cluster_list.append(cluster_data)
            except Exception:
                pass

        return cluster_list

    def get_cluster_list(self, ready: bool = None):
        ls = []
        for cluster_data in self.get_all():
            for provider_type in K8sProviderType:
                if self.exists(cluster_data, provider_type):
                    # handle ready option
                    if ready:
                        kubernetes_cluster = self.select(
                            cluster_data=cluster_data,
                            cluster_provider_type=provider_type,
                        )
                        if not kubernetes_cluster:
                            continue

                        if kubernetes_cluster.ready() != ready:
                            continue

                    # append cluster to list
                    ls.append(cluster_data)
        return ls

    def exists(
        self,
        cluster_data: K8sProviderData,
        cluster_provider_type: K8sProviderType,
    ) -> bool:
        try:
            _ = self.select(
                cluster_data=cluster_data,
                cluster_provider_type=cluster_provider_type,
            )
            return True
        except Exception:
            return False

    def select(
        self,
        cluster_data: K8sProviderData,
        cluster_provider_type: K8sProviderType = settings.UNIKUBE_DEFAULT_PROVIDER_TYPE,
    ) -> Union[AbstractK8sProvider, None]:
        # create config
        config = {
            "id": cluster_data.id,
        }

        if cluster_data.name:
            config["name"] = cluster_data.name

        # get selected kubernetes cluster from factory
        try:
            kubernetes_cluster = kubernetes_cluster_factory.get(
                cluster_provider_type,
                **config,
            )
            return kubernetes_cluster
        except Exception as e:
            console.debug(e)
            return None



from time import sleep

import click
from utils.console import error
from utils.exceptions import UnikubeClusterUnavailableError
from utils.localsystem import K3D, KubeAPI
from utils.project import AppManager, ProjectManager


def retry_command(function):
    def wrapper(*args, **kwargs):
        try:
            res = function(*args, **kwargs)
        except UnikubeClusterUnavailableError:
            error("Cannot reach local cluster.")
            project = ProjectManager().get_active()
            app = AppManager().get_active()
            if click.confirm(f"Should we try to \"project up {project.get('name')}\"?"):
                K3D(project).up(ingress_port=None, workers=None)
                retry_count = 0
                k8s = KubeAPI(project, app)
                while not k8s.is_available and retry_count <= 30:
                    sleep(0.5)
                    retry_count += 1
                if retry_count == 30:
                    error("Could not up project.")
                    exit(1)
                res = function(*args, **kwargs)
            else:
                exit(1)
        return res

    return wrapper



import os
from http.server import BaseHTTPRequestHandler, HTTPServer
from socket import AF_INET, SOCK_STREAM, gethostbyname, socket
from threading import Thread
from urllib.parse import parse_qs

from oic.oic import AccessTokenResponse, AuthorizationResponse, Client

from unikube.authentication.types import AuthenticationData
from unikube.cli import console
from unikube.context import ClickContext

CALLBACK_PORT_RANGE = range(44444, 44448)


def get_callback_port() -> int:
    t_IP = gethostbyname("localhost")
    for port in CALLBACK_PORT_RANGE:
        conn = (s := socket(AF_INET, SOCK_STREAM)).connect_ex((t_IP, port))
        s.close()
        if conn:
            break
    else:
        raise Exception("No port in the range 44444-44447 is available.")
    return port


def run_callback_server(state: str, nonce: str, client: Client, ctx: ClickContext) -> int:
    class CallbackHandler(BaseHTTPRequestHandler):
        """
        This handles the redirect from the Keycloak after the web login.
        A simple http server is started when the user is sent to the keycloak
        web frontend to authenticate.
        """

        def get_post_data(self) -> dict:
            post_body = self.rfile.read(int(self.headers.get("content-length", 0)))
            return {k.decode(): v[0].decode() for k, v in parse_qs(post_body).items()}

        def send_text_response(self, response_body):
            self.send_response(200)
            self.send_header("Content-Type", "text/html")
            self.send_header("Content-Length", str(len(response_body)))
            self.end_headers()
            self.wfile.write(response_body)

        def do_POST(self):
            POST = self.get_post_data()

            if POST["state"] != state:
                raise Exception(f"Invalid state: {POST['state']}")

            response = ctx.auth._get_requesting_party_token(POST["access_token"])

            login_file = open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "login.html"))
            text = login_file.read()
            login_file.close()

            # select response
            if not response["success"]:
                console.error("Login failed!")
                text = (
                    "Login failed! Could not retrieve requesting party token. "
                    "Please try again or contact your System administrator"
                )
            else:
                try:
                    token = ctx.auth.token_from_response(response)
                except Exception as e:
                    console.debug(e)
                    console.debug(response)
                    console.error("Login failed!")
                    text = "Login failed! Your token does not match."
                else:
                    ctx.auth.general_data.authentication = AuthenticationData(
                        email=token["email"],
                        access_token=response["response"]["access_token"],
                        refresh_token=response["response"]["refresh_token"],
                        requesting_party_token=True,
                    )
                    ctx.auth.local_storage_general.set(ctx.auth.general_data)

                    if given_name := token.get("given_name", ""):
                        greeting = f"Hello {given_name}!"
                    else:
                        greeting = "Hello!"

                    html_close = "close"

                    text_html = (
                        f"You have successfully logged in. You can {html_close} this browser tab and return "
                        f"to the shell."
                    )

                    text_plain = (
                        "You have successfully logged in. You can close this browser tab and return to the shell."
                    )

                    greeting_text = "{greeting} {text}".format(greeting=greeting, text=text_plain)

                    greeting_html = "{greeting} {text}".format(greeting=greeting, text=text_html)

                    text = text.replace("##text_placeholder##", greeting_html)
                    text = text.replace("##headline##", "Login Successful")
                    console.success(f"{greeting_text} You are now logged in!")
            response_body = text.encode("utf-8")
            self.send_text_response(response_body)
            Thread(target=server.shutdown).start()

        def log_request(self, *args, **kwargs):
            return

    port = get_callback_port()
    server = HTTPServer(("", port), CallbackHandler)
    Thread(target=server.serve_forever).start()
    return port



import click

import unikube.cli.console as console
from unikube.cli.console.helpers import (
    deck_id_2_display_name,
    organization_id_2_display_name,
    project_id_2_display_name,
)
from unikube.context.helper import convert_context_arguments
from unikube.graphql_utils import GraphQL
from unikube.storage.user import get_local_storage_user


def show_context(ctx, context):
    organization = organization_id_2_display_name(ctx=ctx, id=context.organization_id)
    project = project_id_2_display_name(ctx=ctx, id=context.project_id)
    deck = deck_id_2_display_name(ctx=ctx, id=context.deck_id)

    console.info("Context:")
    console.echo(f"- organization: {organization}")
    console.echo(f"- project: {project}")
    console.echo(f"- deck: {deck}")
    console.echo("")


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def set(ctx, organization=None, project=None, deck=None, **kwargs):
    """
    Set the local context.
    """

    organization_id, project_id, deck_id = convert_context_arguments(
        auth=ctx.auth, organization_argument=organization, project_argument=project, deck_argument=deck
    )

    if not (organization or project or deck):
        organization_id = console.organization_list(ctx=ctx)
        project_id = console.project_list(ctx=ctx, organization_id=organization_id)
        deck_id = console.deck_list(ctx=ctx, organization_id=organization_id, project_id=project_id)
        console.echo("")

    # user_data / context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()

    if organization_id:
        # set organization
        user_data.context.deck_id = None
        user_data.context.project_id = None
        user_data.context.organization_id = organization_id
        local_storage_user.set(user_data)

    if project_id:
        if not organization_id:
            try:
                graph_ql = GraphQL(authentication=ctx.auth)
                data = graph_ql.query(
                    """
                    query($id: UUID) {
                        project(id: $id) {
                            organization {
                                id
                            }
                        }
                    }
                    """,
                    query_variables={
                        "id": project_id,
                    },
                )
                organization_id = data["project"]["organization"]["id"]
            except Exception as e:
                console.debug(e)
                console.exit_generic_error()

        # set project
        user_data.context.deck_id = None
        user_data.context.project_id = project_id
        user_data.context.organization_id = organization_id
        local_storage_user.set(user_data)

    if deck_id:
        if not organization_id or not project_id:
            try:
                graph_ql = GraphQL(authentication=ctx.auth)
                data = graph_ql.query(
                    """
                    query($id: UUID) {
                        deck(id: $id) {
                            project {
                                id
                                organization {
                                    id
                                }
                            }
                        }
                    }
                    """,
                    query_variables={
                        "id": deck_id,
                    },
                )
                organization_id = data["deck"]["project"]["organization"]["id"]
                project_id = data["deck"]["project"]["id"]
            except Exception as e:
                console.debug(e)
                console.exit_generic_error()

        # set deck
        user_data.context.deck_id = deck_id
        user_data.context.project_id = project_id
        user_data.context.organization_id = organization_id
        local_storage_user.set(user_data)

    show_context(ctx=ctx, context=user_data.context)


@click.command()
@click.option("--organization", "-o", is_flag=True, default=False, help="Remove organization context")
@click.option("--project", "-p", is_flag=True, default=False, help="Remove project context")
@click.option("--deck", "-d", is_flag=True, default=False, help="Remove deck context")
@click.pass_obj
def remove(ctx, organization=None, project=None, deck=None, **kwargs):
    """
    Remove the local context.
    """

    # user_data / context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()

    if organization:
        user_data.context.deck_id = None
        user_data.context.project_id = None
        user_data.context.organization_id = None
        local_storage_user.set(user_data)
        console.success("Organization context removed.", _exit=True)

    if project:
        user_data.context.deck_id = None
        user_data.context.project_id = None
        local_storage_user.set(user_data)
        console.success("Project context removed.", _exit=True)

    if deck:
        user_data.context.deck_id = None
        local_storage_user.set(user_data)
        console.success("Deck context removed.", _exit=True)

    # remove complete context
    user_data.context.deck_id = None
    user_data.context.project_id = None
    user_data.context.organization_id = None
    local_storage_user.set(user_data)
    console.success("Context removed.", _exit=True)


@click.command()
@click.pass_obj
def show(ctx, **kwargs):
    """
    Show the context.
    """

    # user_data / context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()

    show_context(ctx=ctx, context=user_data.context)



from tests.login_testcase import LoginTestCase
from unikube.cli.unikube import ps
from unikube.commands import ClickContext


class PsTest(LoginTestCase):
    def test_no_cluster(self):
        result = self.runner.invoke(
            ps,
            obj=ClickContext(),
        )
        self.assertEqual(result.exit_code, 0)



import textwrap
from typing import List

import click

from unikube.cli import console

from .templates import TEMPLATES


def align_spacing(string: str, spaces: int) -> str:
    """
    Prefix multiline string with given number of spaces.

    Just a helper to make the scripts more readable.
    """
    return textwrap.indent(textwrap.dedent(string), " " * spaces)


def get_subcommands(command: dict) -> dict:
    """Extract subcommands from Click's information dictionary retrieved via `to_info_dict`."""
    if "commands" in command:
        commands = command["commands"]
        return {key: get_options(command["commands"][key]) for key in commands.keys()}
    return {}


def get_options(command: dict) -> List[str]:
    """Extract options from Click's information dictionary retrieved via `to_info_dict`."""
    params = command["params"]
    result = []
    for param in params:
        result.extend(filter(lambda x: "--" in x, param["opts"]))
    return result


def render_subcommand_completion(current_command: str, commands: List):
    """Renders case statement for subcommand completion."""
    return """
    ({current_command})
        coms="{commands}"
        COMPREPLY=($(compgen -W "${{coms}}" -- ${{cur}}))
        __ltrim_colon_completions "$cur"
        return 0;
        ;;
    """.format(
        current_command=current_command, commands=" ".join(commands)
    )


def render_options_case(command: str, func_name: str):
    """Renders case statement (function call) for options completion."""
    return align_spacing(
        """
        ({command})
            {func_name}
            ;;""".format(
            command=command, func_name=func_name
        ),
        12,
    )


def render_command_options(command: str, options: List[str]) -> str:
    """Generate bash option string for a command."""
    return """
        ({command})
            opts="${{opts}} {options}"
            ;;""".format(
        command=command, options=" ".join(options)
    )


def render_flag_completion_func(command_dict: dict, name: str) -> (str, str):
    """Renders function for flag completion of command and its subcommands."""
    subcommands = get_subcommands(command_dict)
    options = get_options(command_dict)

    func_name = "__unikube_complete_flags_{name}".format(name=name)

    if subcommands:
        subcommand_cases = align_spacing(
            "\n".join([render_command_options(c[0], c[1]) for c in subcommands.items()]), 16
        )

        return func_name, align_spacing(
            """
        {func_name}() {{
            if [[ $com == $prev ]]; then
                opts="${{opts}} {options}"
            else
                case "$prev" in
                    {subcommand_cases}
                esac
            fi
        }}
        """.format(
                func_name=func_name, options=" ".join(options), subcommand_cases=subcommand_cases
            ),
            0,
        )
    return func_name, align_spacing(
        """
    {func_name}() {{
        opts="${{opts}} {options}"
    }}
    """.format(
            func_name=func_name,
            options=" ".join(options),
        ),
        0,
    )


def render_bash(cli):
    """Renders bash completion script and prints it."""
    with click.Context(cli) as ctx:
        info = ctx.to_info_dict()

        template = TEMPLATES["bash"]

        # static information for rendering
        # could be dynamic in the future (e.g. for aliases)
        function = "_unikube_complete"
        aliases = ["unikube"]
        compdefs = "\n".join(["complete -o default -F {} {}".format(function, alias) for alias in aliases])

        # Based on click's info dict retrieve information about commands and flags.
        # These are then used to render certain parts of the completion.
        commands = info["command"]["commands"].keys()
        command_list = []
        subcommands = []
        functions = []
        for command in commands:
            subs = get_subcommands(info["command"]["commands"][command])
            name, func = render_flag_completion_func(info["command"]["commands"][command], command)
            desc = []
            if name and func:
                functions.append(func)
                desc = [render_options_case(command, name)]

            if len(subs.keys()):
                subcommands.append(render_subcommand_completion(command, list(subs.keys())))

            if len(desc):
                command_list.append("\n".join(desc))

        # Render template with all retrieved information.
        output = template.format(
            function=function,
            flag_complete_functions="\n".join(functions),
            opts=" ".join(sorted([])),
            coms=" ".join(commands),
            command_list="\n".join(command_list),
            compdefs=compdefs,
            subcommands="\n".join(subcommands),
        )

        console.echo(output)


def render_completion_script(cli, shell: str):
    """Renders a completion for a given shell."""
    SUPPORTED_SHELLS = ["bash"]

    if shell not in SUPPORTED_SHELLS:
        console.error(
            "{} is not supported. Following shells are supported: {}.".format(shell, ", ".join(SUPPORTED_SHELLS)),
            _exit=True,
        )

    if shell == "bash":
        render_bash(cli)



import sys
from enum import Enum
from typing import Union

import click_spinner
import requests
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from retrying import retry

import unikube.cli.console as console
from unikube import settings


# EnvironmentType
class EnvironmentType(Enum):
    LOCAL = "LOCAL"
    REMOTE = "REMOTE"


class RetryException(Exception):
    pass


# retry exception logic
def retry_exception(exception):
    return isinstance(exception, RetryException)


class GraphQL:
    def __init__(
        self,
        authentication,
        url=settings.GRAPHQL_URL,
        timeout=settings.GRAPHQL_TIMEOUT,
    ):
        self.url = url
        self.timeout = timeout

        # automatic token refresh
        self.authentication = authentication
        self.access_token = str(authentication.general_data.authentication.access_token)

        # client
        self.client = self._client()

    def _client(self):
        # header
        headers = {
            "Content-type": "application/json",
            "Authorization": "Bearer " + str(self.access_token),
        }

        # transport
        transport = RequestsHTTPTransport(
            url=self.url,
            use_json=True,
            headers=headers,
            verify=False,
            retries=3,
            timeout=self.timeout,
        )

        # client
        client = Client(transport=transport)

        return client

    @retry(retry_on_exception=retry_exception, stop_max_attempt_number=2)
    def query(
        self,
        query: str,
        query_variables: dict = None,
    ) -> Union[dict, None]:
        with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
            try:
                query = gql(query)
                data = self.client.execute(
                    document=query,
                    variable_values=query_variables,
                )

            except requests.exceptions.HTTPError:
                # refresh token
                response = self.authentication.refresh()
                if not response["success"]:
                    console.exit_login_required()

                self.access_token = response["response"]["access_token"]
                self.client = self._client()

                raise RetryException("retry")

        return data



import unittest

from unikube.context.helper import is_valid_uuid4


class IsValidUuid4Test(unittest.TestCase):
    def test_is_valid_uuid4_valid(self):
        result = is_valid_uuid4("51b1d6b3-8375-4859-94f6-73afc05d7275")
        self.assertTrue(result)

    def test_is_valid_uuid4_invalid(self):
        result = is_valid_uuid4("invalid")
        self.assertFalse(result)



from abc import ABC, abstractmethod

from unikube.context.types import ContextData


class UnikubeFileNotFoundError(Exception):
    pass


class UnikubeFileVersionError(Exception):
    pass


class UnikubeFileError(Exception):
    pass


class UnikubeFile(ABC):
    @abstractmethod
    def get_context(self) -> ContextData:
        raise NotImplementedError



import os
from unittest.mock import patch

import pytest
from click.testing import CliRunner

from unikube.cli import auth
from unikube.commands import ClickContext


def test_login_failed():
    runner = CliRunner()
    result = runner.invoke(
        auth.login,
        ["--email", "test@test.de", "--password", "unsecure"],
        obj=ClickContext(),
    )
    assert "[ERROR] Login failed. Please check email and password.\n" in result.output
    assert result.exit_code == 0


def test_login_wrong_token():
    def login(email, password):
        return {"success": True, "response": {"access_token": "WRONG_TOKEN"}}

    runner = CliRunner()
    obj = ClickContext()
    obj.auth.login = login
    result = runner.invoke(
        auth.login,
        ["--email", "test@test.de", "--password", "secure"],
        obj=obj,
    )
    assert "[ERROR] Login failed. Your token does not match." in result.output
    assert result.exit_code == 0


def test_logout():
    runner = CliRunner()
    result = runner.invoke(
        auth.logout,
        obj=ClickContext(),
    )
    assert result.output == "[INFO] Logout completed.\n"
    assert result.exit_code == 0


def test_status_not_logged():
    runner = CliRunner()
    result = runner.invoke(
        auth.status,
        obj=ClickContext(),
    )
    assert result.output == "[INFO] Authentication could not be verified.\n"
    assert result.exit_code == 0


def test_status_success():
    def verify():
        return {"success": True}

    runner = CliRunner()
    obj = ClickContext()
    obj.auth.verify = verify
    result = runner.invoke(
        auth.status,
        obj=obj,
    )
    assert result.output == "[SUCCESS] Authentication verified.\n"
    assert result.exit_code == 0


def test_login_logout_success():
    runner = CliRunner()

    email = os.getenv("TESTRUNNER_EMAIL")
    secret = os.getenv("TESTRUNNER_SECRET")
    assert email is not None
    assert secret is not None

    result = runner.invoke(
        auth.login,
        ["--email", email, "--password", secret],
        obj=ClickContext(),
    )
    assert "[SUCCESS] Login successful. Hello Testrunner!\n" in result.output
    assert result.exit_code == 0

    result = runner.invoke(
        auth.logout,
        obj=ClickContext(),
    )
    assert result.output == "[INFO] Logout completed.\n"
    assert result.exit_code == 0



import click

import unikube.cli.console as console
from unikube.graphql_utils import GraphQL
from unikube.helpers import check_environment_type_local_or_exit, download_manifest
from unikube.local.system import KubeAPI, KubeCtl, Telepresence


def get_deck(ctx, deck_id: str):
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                deck(id: $id) {
                    id
                    title
                    environment {
                        id
                        type
                        valuesPath
                        namespace
                    }
                    project {
                        id
                        title
                        organization {
                            title
                        }
                    }
                }
            }
            """,
            query_variables={"id": deck_id},
        )
        deck = data["deck"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    return deck


def get_cluster(ctx, deck: dict):
    cluster_data = ctx.cluster_manager.get(id=deck["project"]["id"])
    if not cluster_data.name:
        console.error(
            "The project cluster does not exist. Please be sure to run 'unikube project up' first.", _exit=True
        )

    cluster = ctx.cluster_manager.select(cluster_data=cluster_data)

    # check if kubernetes cluster is running/ready
    if not cluster.ready():
        console.error(f"The project cluster for '{cluster.display_name}' is not running.", _exit=True)

    return cluster


def get_ingress_data(deck, provider_data):
    ingresss = KubeAPI(provider_data, deck).get_ingress()
    ingress_data = []
    for ingress in ingresss.items:
        hosts = []
        paths = []
        for rule in ingress.spec.rules:
            hosts.append(f"http://{rule.host}:{provider_data.publisher_port}")  # NOSONAR
            for path in rule.http.paths:
                paths.append(f"{path.path} -> {path.backend.service_name}")
                # this is an empty line in output
            hosts.append("")
            paths.append("")

        ingress_data.append(
            {
                "name": ingress.metadata.name,
                "url": "\n".join(hosts),
                "paths": "\n".join(paths),
            }
        )
    return ingress_data


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.pass_obj
def list(ctx, organization=None, project=None, **kwargs):
    """
    List all decks.
    """

    # context
    organization_id, project_id, _ = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project
    )

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID, $project_id: UUID) {
                allDecks(organizationId: $organization_id, projectId: $project_id) {
                    results {
                        id
                        title
                        project {
                            title
                            organization {
                                title
                            }
                        }
                    }
                }
            }
            """,
            query_variables={
                "organization_id": organization_id,
                "project_id": project_id,
            },
        )
        deck_list = data["allDecks"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    if not deck_list:
        console.warning("No decks available. Please go to https://app.unikube.io and create a project.", _exit=True)

    # format list to table
    table_data = []
    for deck in deck_list:
        data = {}

        if not organization_id:
            data["organization"] = deck["project"]["organization"]["title"]

        if not project_id:
            data["project"] = deck["project"]["title"]

        data["id"] = deck["id"]
        data["title"] = deck["title"]
        table_data.append(data)

    # console
    console.table(data=table_data)


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.argument("deck", required=False)
@click.pass_obj
def info(ctx, organization=None, project=None, deck=None, **kwargs):
    """
    Display further information of the selected deck.
    """

    # context
    organization_id, project_id, deck_id = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project, deck_argument=deck
    )

    # argument
    if not deck_id:
        deck_id = console.deck_list(ctx, organization_id=organization_id, project_id=project_id)
        if not deck_id:
            return None

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                deck(id: $id) {
                    id
                    title
                    description
                    namespace
                    type
                }
            }
            """,
            query_variables={"id": deck_id},
        )
        deck_selected = data["deck"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # console
    if deck_selected:
        console.table(
            data={
                "key": [k for k in deck_selected.keys()],
                "value": [v for v in deck_selected.values()],
            },
            headers=["Key", "Value"],
        )
    else:
        console.error("Deck does not exist.")


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.argument("deck", required=False)
@click.pass_obj
def install(ctx, organization=None, project=None, deck=None, **kwargs):
    """
    Install a deck. For further information please refer to
    :ref:`
    """

    # context
    organization_id, project_id, deck_id = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project, deck_argument=deck
    )

    # argument
    if not deck_id:
        deck_id = console.deck_list(ctx, organization_id=organization_id, project_id=project_id)
        if not deck_id:
            return None

    deck = get_deck(ctx, deck_id=deck_id)

    # cluster
    cluster = get_cluster(ctx=ctx, deck=deck)

    # check environment type
    check_environment_type_local_or_exit(deck=deck)

    # check for switched app
    provider_data = cluster.storage.get()
    telepresence = Telepresence(provider_data)
    if telepresence.intercept_count():
        console.error("It is not possible to install a deck while having an active switch.", _exit=True)

    # download manifest
    general_data = ctx.storage_general.get()
    manifest = download_manifest(
        deck=deck, authentication=ctx.auth, access_token=general_data.authentication.access_token
    )

    # KubeCtl
    provider_data = cluster.storage.get()
    kubectl = KubeCtl(provider_data=provider_data)
    namespace = deck["environment"][0]["namespace"]
    kubectl.create_namespace(namespace)
    with click.progressbar(
        manifest,
        label="[INFO] Installing Kubernetes resources to the cluster.",
    ) as files:
        for file in files:
            kubectl.apply_str(namespace, file["content"])

    # ingress
    ingress_data = get_ingress_data(deck, provider_data)
    if not ingress_data:
        console.info("No ingress configuration available.", _exit=True)

    console.table(
        ingress_data,
        headers={"name": "Name", "url": "URLs", "paths": "Paths"},
    )


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.argument("deck", required=False)
@click.pass_obj
def uninstall(ctx, organization=None, project=None, deck=None, **kwargs):
    """
    Uninstall a deck.
    """

    # context
    organization_id, project_id, deck_id = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project, deck_argument=deck
    )

    # argument
    if not deck_id:
        deck_id = console.deck_list(ctx, organization_id=organization_id, project_id=project_id)
        if not deck_id:
            return None

    deck = get_deck(ctx, deck_id=deck_id)

    # cluster
    cluster = get_cluster(ctx=ctx, deck=deck)

    # check environment type
    check_environment_type_local_or_exit(deck=deck)

    # download manifest
    general_data = ctx.storage_general.get()
    manifest = download_manifest(
        deck=deck, authentication=ctx.auth, access_token=general_data.authentication.access_token
    )

    # KubeCtl
    provider_data = cluster.storage.get()
    kubectl = KubeCtl(provider_data=provider_data)
    namespace = deck["environment"][0]["namespace"]
    with click.progressbar(
        manifest,
        label="[INFO] Deleting Kubernetes resources.",
    ) as files:
        for file in files:
            kubectl.delete_str(namespace, file["content"])

    # console
    console.success("Deck deleted.")


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.argument("deck", required=False)
@click.pass_obj
def ingress(ctx, organization=None, project=None, deck=None, **kwargs):
    """
    Display ingress configuration for *installed* decks. This command prints a table containing URLs, paths and
    the associated backends.
    """

    # context
    organization_id, project_id, deck_id = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project, deck_argument=deck
    )

    # argument
    if not deck_id:
        deck_id = console.deck_list(ctx, organization_id=organization_id, project_id=project_id)
        if not deck_id:
            return None

    deck = get_deck(ctx, deck_id=deck_id)

    # get cluster
    cluster = get_cluster(ctx=ctx, deck=deck)
    provider_data = cluster.storage.get()

    ingress_data = get_ingress_data(deck, provider_data)
    console.table(
        ingress_data,
        headers={"name": "Name", "url": "URLs"},
    )

    if not ingress_data:
        console.warning(
            f"Are you sure the deck is installed? You may have to run 'unikube deck install {deck['title']}' first."
        )



from enum import Enum
from typing import Optional

from pydantic.main import BaseModel


class K8sProviderType(Enum):
    k3d = "k3d"


class K8sProviderData(BaseModel):
    id: str
    name: Optional[str] = None



import os
import re
import subprocess
from time import sleep
from typing import Dict, List, Optional

from semantic_version import Version

import unikube.cli.console as console
from unikube import settings
from unikube.local.providers.abstract_provider import AbstractK8sProvider
from unikube.local.providers.k3d.storage import K3dStorage
from unikube.local.providers.types import K8sProviderType
from unikube.local.system import CMDWrapper


class K3d(AbstractK8sProvider, CMDWrapper):
    kubernetes_cluster_type = K8sProviderType.k3d

    base_command = "k3d"
    _cluster = []

    def __init__(
        self,
        id,
        name: str = None,
        prefix: str = settings.K3D_CLUSTER_PREFIX,
        _debug_output=False,
    ):
        # storage
        storage = K3dStorage(id=id)

        # abstract kubernetes cluster
        AbstractK8sProvider.__init__(
            self,
            id=id,
            name=name,
            storage=storage,
        )

        # CMDWrapper
        self._debug_output = _debug_output

        # cluster name
        cluster_name = prefix + self.name.lower()
        cluster_name = cluster_name.replace(" ", "-")
        self.k3d_cluster_name = cluster_name

    def _clusters(self) -> List[Dict[str, str]]:
        if len(self._cluster) == 0:
            arguments = ["cluster", "list", "--no-headers"]
            process = self._execute(arguments)
            list_output = process.stdout.read()
            clusters = []
            cluster_list = [item.strip() for item in list_output.split("\n")[:-1]]
            for entry in cluster_list:
                cluster = [item.strip() for item in entry.split(" ") if item != ""]
                # todo handle this output
                if len(cluster) != 4:
                    continue
                clusters.append(
                    {
                        "name": cluster[0],
                        "servers": cluster[1],
                        "agents": cluster[2],
                        "loadbalancer": cluster[3] == "true",
                    }
                )
            self._cluster = clusters
        return self._cluster

    def get_kubeconfig(self, wait=10) -> Optional[str]:
        arguments = ["kubeconfig", "get", self.k3d_cluster_name]
        # this is a nasty busy wait, but we don't have another chance
        for i in range(1, wait):
            process = self._execute(arguments)
            if process.returncode == 0:
                break
            else:
                console.info(f"Waiting for the cluster to be ready ({i}/{wait}).")
                sleep(2)

        if process.returncode != 0:
            console.error("Something went completely wrong with the cluster spin up (or we got a timeout).")
        else:
            # we now need to write the kubekonfig to a file
            config = process.stdout.read().strip()
            if not os.path.isdir(os.path.join(settings.CLI_KUBECONFIG_DIRECTORY, self.k3d_cluster_name)):
                os.mkdir(os.path.join(settings.CLI_KUBECONFIG_DIRECTORY, self.k3d_cluster_name))
            config_path = os.path.join(
                settings.CLI_KUBECONFIG_DIRECTORY,
                self.k3d_cluster_name,
                "kubeconfig.yaml",
            )
            file = open(config_path, "w+")
            file.write(config)
            file.close()
            return config_path

    @staticmethod
    def _get_random_unused_port() -> int:
        import socket

        tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tcp.bind(("", 0))
        addr, port = tcp.getsockname()
        tcp.close()
        return port

    def exists(self) -> bool:
        for cluster in self._clusters():
            if cluster["name"] == self.k3d_cluster_name:
                return True
        return False

    def create(
        self,
        ingress_port=None,
        workers=settings.K3D_DEFAULT_WORKERS,
    ):
        v5plus = self.version().major >= 5
        api_port = self._get_random_unused_port()
        if not ingress_port:
            publisher_port = self._get_random_unused_port()
        else:
            publisher_port = ingress_port
        arguments = [
            "cluster",
            "create",
            self.k3d_cluster_name,
            "--agents",
            str(workers),
            "--api-port",
            str(api_port),
            "--port",
            f"{publisher_port}:{settings.K3D_DEFAULT_INGRESS_PORT}@agent{':0' if v5plus else '[0]'}",
            "--servers",
            str(1),
            "--wait",
            "--timeout",
            "120s",
        ]
        self._execute(arguments)

        data = self.storage.get()
        data.name = self.k3d_cluster_name
        data.api_port = api_port
        data.publisher_port = publisher_port
        data.kubeconfig_path = self.get_kubeconfig()
        self.storage.set(data)

        return True

    def start(self):
        arguments = ["cluster", "start", self.k3d_cluster_name]
        p = self._execute(arguments)
        if p.returncode != 0:
            return False
        data = self.storage.get()
        data.kubeconfig_path = self.get_kubeconfig()
        self.storage.set(data)
        return True

    def stop(self):
        arguments = ["cluster", "stop", self.k3d_cluster_name]
        self._execute(arguments)
        return True

    def delete(self):
        arguments = ["cluster", "delete", self.k3d_cluster_name]
        self._execute(arguments)
        self.storage.delete()
        return True

    def version(self) -> Version:
        process = subprocess.run([self.base_command, "--version"], capture_output=True, text=True)
        output = str(process.stdout).strip()
        version_str = re.search(r"(\d+\.\d+\.\d+)", output).group(1)
        return Version(version_str)


class K3dBuilder:
    def __init__(self):
        self._instances = {}

    def __call__(
        self,
        id,
        name=None,
        **_ignored,
    ):
        # get instance from cache
        instance = self._instances.get(id, None)
        if instance:
            return instance

        # create instance
        instance = K3d(
            id,
            name=name,
            prefix=settings.K3D_CLUSTER_PREFIX,
        )
        self._instances[id] = instance

        return instance



from unikube.local.providers.k3d.k3d import K3dBuilder
from unikube.local.providers.types import K8sProviderType


class K8sClusterFactory:
    def __init__(self):
        self._builders = {}

    def register_builder(self, provider_type: K8sProviderType, builder):
        self._builders[provider_type.value] = builder

    def __create(self, provider_type: K8sProviderType, **kwargs):
        builder = self._builders.get(provider_type.value)
        if not builder:
            raise ValueError(provider_type)
        return builder(**kwargs)

    def get(self, provider_type: K8sProviderType, **kwargs):
        return self.__create(provider_type, **kwargs)


kubernetes_cluster_factory = K8sClusterFactory()
kubernetes_cluster_factory.register_builder(K8sProviderType.k3d, K3dBuilder())



import click

import unikube.cli.console as console
from unikube.graphql_utils import GraphQL
from unikube.keycloak.permissions import KeycloakPermissions


@click.command()
@click.pass_obj
def list(ctx, **kwargs):
    """
    List all your organizations.
    """

    _ = ctx.auth.refresh()
    context = ctx.context.get()

    # keycloak
    try:
        keycloak_permissions = KeycloakPermissions(authentication=ctx.auth)
        permission_list = keycloak_permissions.get_permissions_by_scope("organization:*")
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # append "(active)"
    if context.organization_id:
        for permission in permission_list:
            if permission.rsid == context.organization_id:
                permission.rsid += " (active)"

    # console
    organization_list = [
        {
            "id": permission.rsid,
            "name": permission.rsname.replace("organization ", ""),
        }
        for permission in permission_list
    ]
    console.table(
        data=organization_list,
        headers={
            "id": "id",
            "name": "name",
        },
    )


@click.command()
@click.argument("organization", required=False)
@click.pass_obj
def info(ctx, organization, **kwargs):
    """
    Display further information of the selected organization.
    """

    _ = ctx.auth.refresh()

    # context
    organization_id, _, _ = ctx.context.get_context_ids_from_arguments(organization_argument=organization)

    # argument
    if not organization_id:
        organization_id = console.organization_list(ctx)
        if not organization_id:
            return None

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID!) {
                organization(id: $id) {
                    id
                    title
                    description
                }
            }
            """,
            query_variables={"id": organization_id},
        )
        organization_selected = data["organization"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # console
    if organization_selected:
        console.table(
            data={
                "key": [k for k in organization_selected.keys()],
                "value": [v for v in organization_selected.values()],
            },
            headers=["Key", "Value"],
        )
    else:
        console.error("Organization does not exist.")



import click

import unikube.cli.console as console
from unikube.cli.context import show_context
from unikube.graphql_utils import GraphQL
from unikube.local.providers.helper import get_cluster_or_exit
from unikube.local.system import Telepresence
from unikube.storage.user import get_local_storage_user


@click.command()
@click.pass_obj
def ps(ctx, **kwargs):
    """
    Displays the current process state.
    """

    # cluster
    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
    cluster_id_list = [cluster.id for cluster in cluster_list]

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query {
                allProjects {
                    results {
                        title
                        id
                        description
                    }
                }
            }
            """,
        )
        project_list = data["allProjects"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    cluster_data = []
    for project in project_list:
        if project["id"] in cluster_id_list:
            cluster_data.append(project)

    console.info("Project:")
    console.table(
        data={
            "id": [cluster["id"] for cluster in cluster_data],
            "title": [cluster["title"] for cluster in cluster_data],
            "description": [cluster["description"] for cluster in cluster_data],
        },
        headers=["cluster: id", "name", "description"],
    )
    console.echo("")

    # switch
    intercept_count = 0
    if cluster_data:
        cluster = get_cluster_or_exit(ctx, cluster_data[0]["id"])
        provider_data = cluster.storage.get()

        telepresence = Telepresence(provider_data)
        intercept_count = telepresence.intercept_count()

    if intercept_count == 0 or not intercept_count:
        console.info("No app switched!")
    else:
        console.info(f"Apps switched: #{intercept_count}")
    console.echo("")

    # context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()
    show_context(ctx=ctx, context=user_data.context)



import os
import unittest

from click.testing import CliRunner

from unikube.cli import auth
from unikube.commands import ClickContext


class LoginTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()

        email = os.getenv("TESTRUNNER_EMAIL")
        secret = os.getenv("TESTRUNNER_SECRET")
        self.assertIsNotNone(email)
        self.assertIsNotNone(secret)

        self.runner.invoke(
            auth.login,
            ["--email", email, "--password", secret],
            obj=ClickContext(),
        )

    def tearDown(self) -> None:
        result = self.runner.invoke(
            auth.logout,
            obj=ClickContext(),
        )
        self.assertEqual(result.output, "[INFO] Logout completed.\n")
        self.assertEqual(result.exit_code, 0)
